From b33e18da77a73dbfc6882a972e205d31a13703f9 Mon Sep 17 00:00:00 2001
From: Naohiro KORIYAMA <nkoriyama@gmail.com>
Date: Fri, 16 Sep 2011 20:43:16 +0900
Subject: [PATCH] TS demux: #2985 Enhance TS demuxer to support duration and
 current time

---
 modules/demux/ts.c |  324 +++++++++++++++++++++++++++++++++++++++-------------
 1 files changed, 246 insertions(+), 78 deletions(-)

diff --git a/modules/demux/ts.c b/modules/demux/ts.c
index 441fb9b..f94971f 100644
--- a/modules/demux/ts.c
+++ b/modules/demux/ts.c
@@ -336,6 +336,14 @@ struct demux_sys_t
     /* how many TS packet we read at once */
     int         i_ts_read;
 
+    /* to determine length and time */
+    int         i_pid_first_pcr;
+    mtime_t     i_first_pcr;
+    mtime_t     i_current_pcr;
+    mtime_t     i_last_pcr;
+    mtime_t     i_time;
+    int         i_mux_rate;
+
     /* All pid */
     ts_pid_t    pid[8192];
 
@@ -401,6 +409,10 @@ static inline int PIDGet( block_t *p )
 
 static bool GatherPES( demux_t *p_demux, ts_pid_t *pid, block_t *p_bk );
 
+static block_t* ReadTSPacket( demux_t *p_demux );
+static void GetPCR( block_t *p_pkt, mtime_t *p_pcr );
+static void GetFirstPCR( demux_t *p_demux );
+static void GetLastPCR( demux_t *p_demux );
 static void PCRHandle( demux_t *p_demux, ts_pid_t *, block_t * );
 
 static iod_descriptor_t *IODNew( int , uint8_t * );
@@ -788,6 +800,18 @@ static int Open( vlc_object_t *p_this )
     p_sys->b_silent = var_CreateGetBool( p_demux, "ts-silent" );
     p_sys->b_split_es = var_InheritBool( p_demux, "ts-split-es" );
 
+    bool can_seek = false;
+    stream_Control( p_demux->s, STREAM_CAN_SEEK, &can_seek );
+    if( can_seek )
+    {
+        GetFirstPCR( p_demux );
+        GetLastPCR( p_demux );
+        if( p_sys->i_last_pcr - p_sys->i_first_pcr > 0 )
+        {
+            p_sys->i_mux_rate = (stream_Size( p_demux->s )) * INT64_C(1000000) / ((p_sys->i_last_pcr - p_sys->i_first_pcr) * 100 / 9);
+        }
+    }
+
     while( !p_sys->b_file_out && p_sys->i_pmt_es <= 0 &&
            vlc_object_alive( p_demux ) )
     {
@@ -1013,71 +1037,81 @@ static int DemuxFile( demux_t *p_demux )
     return 1;
 }
 
-/*****************************************************************************
- * Demux:
- *****************************************************************************/
-static int Demux( demux_t *p_demux )
+static block_t* ReadTSPacket( demux_t *p_demux )
 {
     demux_sys_t *p_sys = p_demux->p_sys;
-    bool b_wait_es = p_sys->i_pmt_es <= 0;
 
-    /* We read at most 100 TS packet or until a frame is completed */
-    for( int i_pkt = 0; i_pkt < p_sys->i_ts_read; i_pkt++ )
-    {
-        bool         b_frame = false;
-        block_t     *p_pkt;
+    block_t     *p_pkt;
 
-        /* Get a new TS packet */
-        if( !( p_pkt = stream_Block( p_demux->s, p_sys->i_packet_size ) ) )
-        {
-            msg_Dbg( p_demux, "eof ?" );
-            return 0;
-        }
+    /* Get a new TS packet */
+    if( !( p_pkt = stream_Block( p_demux->s, p_sys->i_packet_size ) ) )
+    {
+        msg_Dbg( p_demux, "eof ?" );
+        return NULL;
+    }
 
-        /* Check sync byte and re-sync if needed */
-        if( p_pkt->p_buffer[0] != 0x47 )
+    /* Check sync byte and re-sync if needed */
+    if( p_pkt->p_buffer[0] != 0x47 )
+    {
+        msg_Warn( p_demux, "lost synchro" );
+        block_Release( p_pkt );
+        while( vlc_object_alive (p_demux) )
         {
-            msg_Warn( p_demux, "lost synchro" );
-            block_Release( p_pkt );
+            const uint8_t *p_peek;
+            int i_peek, i_skip = 0;
 
-            while( vlc_object_alive (p_demux) )
+            i_peek = stream_Peek( p_demux->s, &p_peek,
+                    p_sys->i_packet_size * 10 );
+            if( i_peek < p_sys->i_packet_size + 1 )
             {
-                const uint8_t *p_peek;
-                int i_peek, i_skip = 0;
-
-                i_peek = stream_Peek( p_demux->s, &p_peek,
-                                      p_sys->i_packet_size * 10 );
-                if( i_peek < p_sys->i_packet_size + 1 )
-                {
-                    msg_Dbg( p_demux, "eof ?" );
-                    return 0;
-                }
+                msg_Dbg( p_demux, "eof ?" );
+                return NULL;
+            }
 
-                while( i_skip < i_peek - p_sys->i_packet_size )
-                {
-                    if( p_peek[i_skip] == 0x47 &&
+            while( i_skip < i_peek - p_sys->i_packet_size )
+            {
+                if( p_peek[i_skip] == 0x47 &&
                         p_peek[i_skip + p_sys->i_packet_size] == 0x47 )
-                    {
-                        break;
-                    }
-                    i_skip++;
-                }
-
-                msg_Dbg( p_demux, "skipping %d bytes of garbage", i_skip );
-                stream_Read( p_demux->s, NULL, i_skip );
-
-                if( i_skip < i_peek - p_sys->i_packet_size )
                 {
                     break;
                 }
+                i_skip++;
             }
+            msg_Dbg( p_demux, "skipping %d bytes of garbage", i_skip );
+            stream_Read( p_demux->s, NULL, i_skip );
 
-            if( !( p_pkt = stream_Block( p_demux->s, p_sys->i_packet_size ) ) )
+            if( i_skip < i_peek - p_sys->i_packet_size )
             {
-                msg_Dbg( p_demux, "eof ?" );
-                return 0;
+                break;
             }
         }
+        if( !( p_pkt = stream_Block( p_demux->s, p_sys->i_packet_size ) ) )
+        {
+            msg_Dbg( p_demux, "eof ?" );
+            return NULL;
+        }
+    }
+    return p_pkt;
+}
+
+/*****************************************************************************
+ * Demux:
+ *****************************************************************************/
+static int Demux( demux_t *p_demux )
+{
+    demux_sys_t *p_sys = p_demux->p_sys;
+    bool b_wait_es = p_sys->i_pmt_es <= 0;
+
+    /* We read at most 100 TS packet or until a frame is completed */
+    for( int i_pkt = 0; i_pkt < p_sys->i_ts_read; i_pkt++ )
+    {
+        bool         b_frame = false;
+        block_t     *p_pkt;
+        if( !(p_pkt = ReadTSPacket( p_demux )) )
+        {
+            msg_Dbg( p_demux, "read ts packet error" );
+            break;
+        }
 
         if( p_sys->b_start_record )
         {
@@ -1219,41 +1253,38 @@ static int Control( demux_t *p_demux, int i_query, va_list args )
             return VLC_EGENERIC;
 
         return VLC_SUCCESS;
-#if 0
 
     case DEMUX_GET_TIME:
         pi64 = (int64_t*)va_arg( args, int64_t * );
-        if( p_sys->i_time < 0 )
+        if( (p_sys->b_dvb_meta && p_sys->b_access_control) || p_sys->i_time <= 0 )
         {
-            *pi64 = 0;
-            return VLC_EGENERIC;
+            if( DVBEventInformation( p_demux, pi64, NULL ) )
+            {
+                *pi64 = 0;
+            }
+        }
+        else
+        {
+            *pi64 = p_sys->i_time;
         }
-        *pi64 = p_sys->i_time;
         return VLC_SUCCESS;
 
     case DEMUX_GET_LENGTH:
         pi64 = (int64_t*)va_arg( args, int64_t * );
-        if( p_sys->i_mux_rate > 0 )
+        if( (p_sys->b_dvb_meta && p_sys->b_access_control) || p_sys->i_mux_rate <= 0 )
+        {
+            if( DVBEventInformation( p_demux, NULL, pi64 ) )
+            {
+                *pi64 = 0;
+            }
+        }
+        else
         {
-            *pi64 = INT64_C(1000000) * ( stream_Size( p_demux->s ) / 50 ) /
+            *pi64 = INT64_C(1000000) * ( stream_Size( p_demux->s ) ) /
                     p_sys->i_mux_rate;
-            return VLC_SUCCESS;
         }
-        *pi64 = 0;
-        return VLC_EGENERIC;
-#else
-    case DEMUX_GET_TIME:
-        pi64 = (int64_t*)va_arg( args, int64_t * );
-        if( DVBEventInformation( p_demux, pi64, NULL ) )
-            *pi64 = 0;
         return VLC_SUCCESS;
 
-    case DEMUX_GET_LENGTH:
-        pi64 = (int64_t*)va_arg( args, int64_t * );
-        if( DVBEventInformation( p_demux, NULL, pi64 ) )
-            *pi64 = 0;
-        return VLC_SUCCESS;
-#endif
     case DEMUX_SET_GROUP:
     {
         vlc_list_t *p_list;
@@ -1847,24 +1878,161 @@ static void ParsePES( demux_t *p_demux, ts_pid_t *pid )
     }
 }
 
-static void PCRHandle( demux_t *p_demux, ts_pid_t *pid, block_t *p_bk )
+static void GetPCR( block_t *p_pkt, mtime_t *p_pcr )
 {
-    demux_sys_t   *p_sys = p_demux->p_sys;
-    const uint8_t *p = p_bk->p_buffer;
+    const uint8_t *p = p_pkt->p_buffer;
 
-    if( p_sys->i_pmt_es <= 0 )
-        return;
+    *p_pcr = 0;
 
     if( ( p[3]&0x20 ) && /* adaptation */
         ( p[5]&0x10 ) &&
         ( p[4] >= 7 ) )
     {
         /* PCR is 33 bits */
-        const mtime_t i_pcr = ( (mtime_t)p[6] << 25 ) |
-                              ( (mtime_t)p[7] << 17 ) |
-                              ( (mtime_t)p[8] << 9 ) |
-                              ( (mtime_t)p[9] << 1 ) |
-                              ( (mtime_t)p[10] >> 7 );
+        *p_pcr = ( (mtime_t)p[6] << 25 ) |
+                 ( (mtime_t)p[7] << 17 ) |
+                 ( (mtime_t)p[8] << 9 ) |
+                 ( (mtime_t)p[9] << 1 ) |
+                 ( (mtime_t)p[10] >> 7 );
+    }
+}
+
+static void GetFirstPCR( demux_t *p_demux )
+{
+    demux_sys_t *p_sys = p_demux->p_sys;
+
+    int64_t i_current_pos = stream_Tell( p_demux->s );
+
+    while( vlc_object_alive (p_demux) )
+    {
+        block_t     *p_pkt;
+        if( !(p_pkt = ReadTSPacket( p_demux )) )
+        {
+            msg_Dbg( p_demux, "read ts packet error" );
+            break;
+        }
+
+        /* Parse the TS packet */
+        ts_pid_t *p_pid = &p_sys->pid[PIDGet( p_pkt )];
+
+        bool b_pcr = false;
+        if( p_pid->b_valid )
+        {
+            if( p_pid->psi )
+            {
+                if( p_pid->i_pid == 0 || ( p_sys->b_dvb_meta && ( p_pid->i_pid == 0x11 || p_pid->i_pid == 0x12 || p_pid->i_pid == 0x14 ) ) )
+                {
+                    dvbpsi_PushPacket( p_pid->psi->handle, p_pkt->p_buffer );
+                }
+                else
+                {
+                    for( int i_prg = 0; i_prg < p_pid->psi->i_prg; i_prg++ )
+                    {
+                        dvbpsi_PushPacket( p_pid->psi->prg[i_prg]->handle,
+                                           p_pkt->p_buffer );
+                    }
+                }
+            }
+            else
+            {
+                b_pcr = true;
+            }
+        }
+        else
+        {
+            b_pcr = true;
+        }
+        if( !b_pcr )
+        {
+            block_Release( p_pkt );
+            continue;
+        }
+
+        if( p_sys->i_pmt_es > 0 )
+        {
+            if( p_sys->i_pid_first_pcr == 0  )
+            {
+                mtime_t i_pcr;
+                GetPCR( p_pkt, &i_pcr );
+                if( i_pcr > 0 )
+                {
+                    p_sys->i_pid_first_pcr = p_pid->i_pid;
+                    p_sys->i_first_pcr = i_pcr;
+                }
+            }
+        }
+        block_Release( p_pkt );
+        
+        if( p_sys->i_first_pcr > 0 )
+            break;
+    }
+
+    stream_Seek( p_demux->s, i_current_pos );
+}
+
+static void GetLastPCR( demux_t *p_demux )
+{
+    demux_sys_t *p_sys = p_demux->p_sys;
+
+    int64_t i_current_pos = stream_Tell( p_demux->s );
+    int64_t i_last_pos = stream_Size( p_demux->s ) - p_sys->i_packet_size;
+    int64_t i_pos = i_last_pos - p_sys->i_packet_size * 1000; /* FIXME how many packets to get last pcr? */
+    if( i_pos < 0 )
+        return;
+
+    if( stream_Seek( p_demux->s, i_pos ) )
+        return;
+
+    while( vlc_object_alive( p_demux ) )
+    {
+        block_t     *p_pkt;
+        if( !(p_pkt = ReadTSPacket( p_demux )) )
+        {
+            msg_Dbg( p_demux, "read ts packet error" );
+            break;
+        }
+
+        ts_pid_t *p_pid = &p_sys->pid[PIDGet( p_pkt )];
+        if( p_sys->i_pmt_es > 0 )
+        {
+            if( p_sys->i_pid_first_pcr ==  p_pid->i_pid )
+            {
+                mtime_t i_pcr;
+                GetPCR( p_pkt, &i_pcr );
+                if( i_pcr > 0 )
+                {
+                    p_sys->i_last_pcr = i_pcr;
+                }
+            }
+        }
+        block_Release( p_pkt );
+        if( stream_Tell( p_demux->s ) >= i_last_pos )
+            break;
+    }
+
+    stream_Seek( p_demux->s, i_current_pos );
+}
+
+static void PCRHandle( demux_t *p_demux, ts_pid_t *pid, block_t *p_bk )
+{
+    demux_sys_t   *p_sys = p_demux->p_sys;
+
+    if( p_sys->i_pmt_es <= 0 )
+        return;
+
+    mtime_t i_pcr;
+    GetPCR( p_bk, &i_pcr );
+    if( i_pcr > 0 )
+    {
+        if( p_sys->i_pid_first_pcr == pid->i_pid )
+        {
+            p_sys->i_current_pcr = i_pcr;
+
+            if( p_sys->i_current_pcr > 0 && p_sys->i_first_pcr > 0 && p_sys->i_current_pcr > p_sys->i_first_pcr )
+            {
+                p_sys->i_time = (p_sys->i_current_pcr - p_sys->i_first_pcr) * 100 / 9;
+            }
+        }
 
         /* Search program and set the PCR */
         for( int i = 0; i < p_sys->i_pmt; i++ )
-- 
1.7.4.4

